@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: execute.c,v 1.107 2008-07-11 14:13:02 timo Exp $ */
+/* $Id: execute.c,v 1.118 2011-04-10 10:12:09 timo Exp $ */
 
 #include "ffe.h"
 #include <stdlib.h>
@@ -45,6 +45,10 @@ extern struct replace *replace;
 struct input_file *files = NULL;
 static struct input_file *current_file = NULL;
 static FILE *input_fp = NULL;
+static int ungetchar = -1;
+static char *default_output_file = NULL;
+static FILE *default_output_fp = NULL;
+
 static char *output_file = NULL;
 static FILE *output_fp = NULL;
 
@@ -74,7 +78,7 @@ static long long int current_offset = 0;
 static long long int current_file_offset = 0;
 
 /* examples of non matching lines */
-#define NO_MATCH_LINES 10
+#define NO_MATCH_LINES 1
 static int no_matching_lines = 0;
 
 /* header definition from structure */
@@ -86,8 +90,14 @@ static char debug_file[128];
 static FILE *debug_fp = NULL;
 static long int debug_lineno = 0;
 
-static uint8_t bcd_to_ascii[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','\000'};
-static uint8_t hex_to_ascii[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
+static uint8_t bcd_to_ascii_cap[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','\000'};
+static uint8_t hex_to_ascii_cap[]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
+
+static uint8_t bcd_to_ascii_low[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','\000'};
+static uint8_t hex_to_ascii_low[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
+
+static uint8_t *bcd_to_ascii;
+static uint8_t *hex_to_ascii;
 
 inline uint8_t
 htocl(uint8_t hex)
@@ -119,21 +129,43 @@ set_output_file(char *name)
 {
     if(name == NULL)
     {
-        output_fp = stdout;
-        output_file = "(stdout)";
+        default_output_fp = stdout;
+        default_output_file = "(stdout)";
     } else
     {
-        output_fp = xfopen(name,"w");
-        output_file = name;
+        default_output_fp = xfopen(name,"w");
+        default_output_file = name;
     }
+    output_fp = default_output_fp;
+    output_file = default_output_file;
 }
 
 void 
 close_output_file()
 {
-    if(fclose(output_fp) != 0)
+    struct output *o = output;
+    int stdoutclosed = 0;
+
+    if(default_output_fp == stdout) stdoutclosed = 1;
+    if(fclose(default_output_fp) != 0)
     {
-        panic("Error closing file",output_file,strerror(errno));
+        panic("Error closing file",default_output_file,strerror(errno));
+    }
+
+    while(o != NULL)
+    {
+        if(o->ofp != NULL)
+        {
+            if((o->ofp == stdout && !stdoutclosed) || o->ofp != stdout)
+            {
+                if(o->ofp == stdout) stdoutclosed = 1;
+                if(fclose(o->ofp) != 0)
+                {
+                    panic("Error closing file",o->output_file,strerror(errno));
+                }
+            }
+        }
+        o = o->next;
     }
 }
 
@@ -158,6 +190,64 @@ set_input_file(char *name)
     f->lineno = 0;
 }
 
+static FILE *
+open_input_stream(char *file,char type)
+{
+    int fds[2];
+    pid_t pid;
+    FILE *ret = NULL;
+    char command[1024];
+
+    if(ffe_open != NULL && ffe_open[0] != '\000')                // use preprocessor
+    {
+#if defined(HAVE_WORKING_FORK) && defined(HAVE_DUP2) && defined(HAVE_PIPE)
+       sprintf(command,ffe_open,file);
+       if (pipe(fds) != 0) panic("Cannot create pipe",strerror(errno),NULL);
+       pid = fork();
+       if(pid == (pid_t) 0) /* Child */
+       {
+          close(fds[0]);
+          if(dup2(fds[1],STDOUT_FILENO) == -1) panic("dup2 error",strerror(errno),NULL);
+          if(execl(SHELL_CMD, "sh", "-c", command, NULL) == -1) panic("Starting a shell with execl failed",command,strerror(errno));
+          close(fds[1]);
+          _exit(EXIT_SUCCESS);
+       } else if(pid > (pid_t) 0)
+       {
+          close(fds[1]);
+          ret = fdopen(fds[0],"r");
+          if(ret == NULL) panic("Cannot read from command",command,strerror(errno));
+          
+          ungetchar = fgetc(ret);
+
+          if(ungetchar == EOF)       // check if pipe returns something, if not open file normally
+          {
+              ungetchar = -1;
+              fclose(ret);
+              ret = NULL;
+          }
+       } else
+       {
+          panic("Cannot fork",strerror(errno),NULL);
+       }
+#else
+       panic("Input preprocessing is not supported in this system",NULL,NULL);
+#endif
+    }
+
+    if(ret == NULL)
+    {
+        if(type == BINARY)
+        {
+            ret = xfopenb(file,"r");
+        } else
+        {
+            ret = xfopen(file,"r");
+        }
+    }
+    return ret;
+}
+
+
 void
 open_input_file(int stype)
 {
@@ -170,13 +260,7 @@ open_input_file(int stype)
         files->name = "(stdin)";
     } else
     {
-        if(stype == BINARY)
-        {
-            input_fp = xfopenb(files->name,"r");
-        } else
-        {
-            input_fp = xfopen(files->name,"r");
-        }
+        input_fp = open_input_stream(files->name,stype);
     }
     current_file = files;
     current_file->lineno = 0;
@@ -202,6 +286,65 @@ complete_line_in_bin_buffer(int *ccount)
 }
 
 
+/* read from input stream. 
+   Check if ungetchar contains a valid char and write it to buffer then read nmenb - 1 chars
+   NOTE! it is assumed that size == 1...
+*/
+static size_t
+uc_fread(uint8_t *ptr, size_t size, size_t nmemb, FILE *stream)
+{
+    size_t ret = 0;
+
+    if(ungetchar != -1)  // there is peeked char in ungetchar, write it to buffer and read the rest
+    {
+        *ptr = (uint8_t) ungetchar;
+        ungetchar = -1;
+        ptr++;
+        nmemb--;
+        ret = 1;
+    }
+
+    ret += fread(ptr,size,nmemb,stream);
+    return ret;
+}
+
+/* read line from input stream. 
+   Check if ungetchar contains a valid char and write it to buffer and then read the rest
+*/
+static int 
+uc_fgets(uint8_t *s, int size, FILE *stream)
+{
+    int ret;
+    char *orig_s = s;
+
+    if(ungetchar != -1)  // there is peeked char in ungetchar, write it to buffer and read the rest
+    {
+        *s = (uint8_t) ungetchar;
+        ungetchar = -1;
+        if(*s == '\n')
+        {
+            s++;
+            *s = '\000';
+            return 1;
+        }
+        s++;
+        size--;
+    } 
+
+    if(fgets(s,size,stream) != NULL)
+    {
+        ret = strlen(orig_s);
+    } else
+    {
+        ret = -1;
+    }
+
+    return ret;
+}
+
+
+
+
 /* reads one file from input */
 /* returns the line length */
 /* return -1 on EOF */
@@ -223,7 +366,7 @@ read_input_line(int stype)
                 memmove(read_buffer,read_buffer+last_consumed,read_buffer_size - last_consumed);
                 if(!eof)
                 {
-                    ccount = fread(read_buffer + (read_buffer_size - last_consumed),1,last_consumed,input_fp);
+                    ccount = uc_fread(read_buffer + (read_buffer_size - last_consumed),1,last_consumed,input_fp);
                     if(ccount < last_consumed) eof = 1;
                     ccount += read_buffer_size - last_consumed; // number of usable octets
                 } else
@@ -238,7 +381,7 @@ read_input_line(int stype)
                     current_file_lineno = current_file->lineno;
                     return ccount;  
                 }
-                ccount = fread(read_buffer,1,read_buffer_size,input_fp);
+                ccount = uc_fread(read_buffer,1,read_buffer_size,input_fp);
                 if(ccount < read_buffer_size) eof = 1;
             }
 
@@ -279,8 +422,8 @@ read_input_line(int stype)
                     len = (int) (bbuffer - read_buffer);
                 }
                 
-                bbuffer++;
-                 
+                if ((int) (bbuffer - read_buffer) < ccount) bbuffer++;
+
                 ccount -= (int) (bbuffer - read_buffer);
 
                 current_file_lineno = current_file->lineno;
@@ -290,16 +433,9 @@ read_input_line(int stype)
                 return len ;
             }
 
-#ifdef HAVE_GETLINE
-            ccount = getline((char **) &read_buffer,&read_buffer_size,input_fp);
-#else
-            if(!eof && (fgets(read_buffer,read_buffer_size,input_fp) != NULL))
-            {
-                ccount = strlen(read_buffer);
-            } else
-            {
-                ccount = -1;
-            }
+            ccount = uc_fgets(read_buffer,read_buffer_size,input_fp);
+#ifdef WIN32
+            if(eof)  ccount = -1;
 #endif
         }
 
@@ -318,16 +454,13 @@ read_input_line(int stype)
                     current_file->name = "(stdin)";
                 } else
                 {
+                    input_fp = open_input_stream(current_file->name,stype);
                     if(stype == BINARY)
                     {
-                        input_fp = xfopenb(current_file->name,"r");
                         last_consumed = 0;
                         current_file_offset = 0;
-                        eof = 0;
-                    } else
-                    {
-                        input_fp = xfopen(current_file->name,"r");
-                    }
+                    } 
+                    eof = 0;
                 }
                 current_file_name = current_file->name;
                 current_file->lineno = 0;
@@ -480,7 +613,7 @@ int
 vote_record(uint8_t quote,char *type,int header,struct record *record,int length,uint8_t *buffer)
 {
     register struct id *i = record->i;
-    int vote = 0;
+    int vote = 0,len;
     int ids = 0;
 
     while(i != NULL)
@@ -535,11 +668,13 @@ vote_record(uint8_t quote,char *type,int header,struct record *record,int length
         switch(type[0])
         {
             case FIXED_LENGTH:
-                if((record->arb_length && record->length <= length) ||
-                   (!record->arb_length && record->length == length)) vote++;
+                if((record->arb_length == RL_MIN && record->length <= length) ||
+                   (record->arb_length == RL_STRICT && record->length == length)) vote++;
                 break;
             case SEPARATED:
-                if(record->length == get_field_count(quote,type,buffer)) vote++;
+                len = get_field_count(quote,type,buffer);
+                if((record->arb_length == RL_STRICT && record->length == len)  ||
+                   (record->arb_length == RL_MIN && record->length <= len)) vote++;
                 break;
             case BINARY:
                 if(((vote == ids && ids) || !ids) && record->length <= length) vote++; /* exact binary length cannot be checked */
@@ -773,11 +908,11 @@ flush_write()
 void
 print_raw(int size, uint8_t *buffer,int stype)
 {
-    if(fwrite(buffer,1,size,output_fp) != size)
+    if(fwrite(buffer,1,size,default_output_fp) != size)
     {
-        panic("Error writing to",output_file,NULL);
+        panic("Error writing to",default_output_file,NULL);
     }
-    if(stype != BINARY) fputc('\n',output_fp);
+    if(stype != BINARY) fputc('\n',default_output_fp);
 }
     
 
@@ -828,6 +963,12 @@ print_text(struct structure *s, struct record *r,uint8_t *buffer)
                     sprintf(num,"%lld",current_file_offset);
                     writes(num);
                     break;
+                case 'g':
+                    if(r->level && r->level->group_name) writes(r->level->group_name);
+                    break;
+                case 'n':
+                    if(r->level && r->level->element_name) writes(r->level->element_name);
+                    break;
                 case '%':
                     writec('%');
                     break;
@@ -965,7 +1106,7 @@ update_field_positions(char *type,uint8_t quote,int arbit_len,struct field *fiel
         case BINARY:
             break;
         case FIXED_LENGTH:
-            if(arbit_len)
+            if(arbit_len == RL_MIN)
             {
                 while(f != NULL)
                 {
@@ -1053,6 +1194,7 @@ print_fixed_field(uint8_t format,struct field *f,uint8_t *buffer)
     if(f->const_data != NULL)
     {
         data = f->const_data;
+        if (format == 't') format = 'd';  // Dont't trim consts
     } else
     {
         if(f->bposition < 0) return;  /* last variable length field is missing */
@@ -1063,6 +1205,7 @@ print_fixed_field(uint8_t format,struct field *f,uint8_t *buffer)
     {
         case 'd':
         case 'D':
+        case 'C':
         case 'e':
         case 'x':
             if(f->length)
@@ -1140,6 +1283,7 @@ print_binary_field(uint8_t format,struct field *f,uint8_t *buffer)
         case 'd':
         case 't':
         case 'D':
+        case 'C':
         case 'x':
             switch(f->type)
             {
@@ -1278,7 +1422,7 @@ print_separated_field(uint8_t format,uint8_t quote,uint8_t separator,struct fiel
     {
         if(f->bposition < 0)
         {
-            if(format == 'D') while(write_pos - start < f->length) writec(' ');
+            if(format == 'D' || format == 'C') while(write_pos - start < f->length) writec(' ');
             return;
         }
 
@@ -1287,10 +1431,11 @@ print_separated_field(uint8_t format,uint8_t quote,uint8_t separator,struct fiel
             case 'd':
             case 't':
             case 'D':
+            case 'C':
             case 'e':
             case 'x':
                 p = &buffer[f->bposition];
-                if(quote || format == 't') while(isspace(*p)) p++;
+                if(quote || format == 't') while(*p != separator && isspace(*p)) p++;
                 if(*p == quote && quote) 
                 {
                     p++;
@@ -1312,10 +1457,12 @@ print_separated_field(uint8_t format,uint8_t quote,uint8_t separator,struct fiel
                                 while(write_pos > start && isspace(*write_pos)) write_pos--;
                                 write_pos++;
                             }
-                            if(format == 'D') while(write_pos - start < f->length) writec(' ');
+                            if(format == 'D' || format == 'C') while(write_pos - start == f->length) writec(' ');
                             return;
                         }
                     }
+
+                    if(format == 'C' && (write_pos - start) == f->length) return;
                     writec(*p);
                     if(*p) p++;
                 }            
@@ -1325,7 +1472,7 @@ print_separated_field(uint8_t format,uint8_t quote,uint8_t separator,struct fiel
                     while(write_pos > start && isspace(*write_pos)) write_pos--;
                     write_pos++;
                 }
-                if(format == 'D') while(write_pos - start < f->length) writec(' ');
+                if(format == 'D' || format == 'C') while(write_pos - start < f->length) writec(' ');
                 break;
         }
     }
@@ -1387,6 +1534,14 @@ make_lookup(struct lookup *l,uint8_t *search)
     return ret_val;
 }
 
+void
+print_indent(uint8_t *buffer,int times)
+{
+    start_write();
+    while(times--) writes(buffer);
+    flush_write();
+}
+
 
 /* print fields */
 /* returns the count of fields actually printed */
@@ -1435,6 +1590,16 @@ print_fields(struct structure *s, struct record *r,uint8_t *buffer)
 
             if(pf->f->lookup != NULL) d = o->lookup;
 
+            if(o->hex_cap)
+            {
+                bcd_to_ascii = bcd_to_ascii_cap;
+                hex_to_ascii = hex_to_ascii_cap;
+            } else
+            {
+                bcd_to_ascii = bcd_to_ascii_low;
+                hex_to_ascii = hex_to_ascii_low;
+            }
+
             while(*d)
             {
                 if(justify == *d)
@@ -1534,6 +1699,7 @@ print_fields(struct structure *s, struct record *r,uint8_t *buffer)
                     case 'd':
                     case 't':
                     case 'D':
+                    case 'C':
                     case 'e':
                     case 'x':
                         field_start = write_pos;
@@ -1594,7 +1760,7 @@ print_fields(struct structure *s, struct record *r,uint8_t *buffer)
                     d = rep_pos;
                     d++;
                     replacing = 0;
-                    if(*rep_pos == 'D')
+                    if(*rep_pos == 'D' || *rep_pos == 'C')
                     {
                         if(write_pos - rep_start < pf->f->length)
                         {
@@ -1655,8 +1821,16 @@ print_fields(struct structure *s, struct record *r,uint8_t *buffer)
             {
                 if(indent != NULL)
                 {
-                    fputs(indent,output_fp);
-                    fputs(indent,output_fp);
+                    if(r->level)
+                    {
+                        int i;
+                        i = get_indent_depth(r->level->level);
+                        while(i--) fputs(indent,output_fp);
+                    } else
+                    {
+                        fputs(indent,output_fp);
+                        fputs(indent,output_fp);
+                    }
                 }
                 if((justify != LEFT_JUSTIFY && justify != RIGHT_JUSTIFY && max_justify_len) || justify == RIGHT_JUSTIFY)
                 {
@@ -1811,7 +1985,6 @@ init_structure(struct structure *s,struct record *current_record,int length, uin
                 if(strcasecmp(f->name,e->field) == 0)
                 {
                     e->found = 1;
-                    f->exp = e;
                 }
                 e = e->next;
             }
@@ -1873,11 +2046,121 @@ invalid_input(char *file, long int lineno,int strict,int length,int stype)
     if(strict) panic("Using option -l does not cause program to abort in case of invalid input",NULL,NULL);
 }
 
+
+static int
+hash_scan_expression(struct expr_list **list,char *value,int casecmp)
+{
+	register struct expr_list *l = list[hash(value,0)];
+
+	while(l != NULL)
+	{
+		if(casecmp)
+		{
+			if(strcasecmp(l->value,value) == 0) return 1;
+		} else
+		{
+			if(strcmp(l->value,value) == 0) return 1;
+		}
+                l = l->next;
+	}
+        return 0;
+} 
+    
+static int
+hash_scan_expression_len(struct expr_list **list,char *value,int casecmp,size_t comp_len)
+{
+	register struct expr_list *l = list[hash(value,comp_len)];
+
+	while(l != NULL)
+	{
+		if(casecmp)
+		{
+			if(strncasecmp(l->value,value,l->value_len) == 0) return 1;
+		} else
+		{
+			if(strncmp(l->value,value,l->value_len) == 0) return 1;
+		}
+                l = l->next;
+	}
+        return 0;
+} 
+
+static inline
+scan_expression_list(struct expr_list *l,char *value,char op,int casecmp)
+{
+    while(l != NULL)
+    {
+        switch(op)
+        {
+            case OP_START:
+                if(casecmp)
+                {
+                    if(strncasecmp(l->value,value,l->value_len) == 0) return 1;
+                } else
+                {
+                    if(strncmp(l->value,value,l->value_len) == 0) return 1;
+                }
+                break;
+            case OP_EQUAL:
+            case OP_NOT_EQUAL:
+                if(casecmp)
+                {
+                    if(strcasecmp(l->value,value) == 0) return 1;
+                } else
+                {
+                    if(strcmp(l->value,value) == 0) return 1;
+                }
+                break;
+            case OP_CONTAINS:
+                if(casecmp)
+                {
+                    if(strcasestr(value,l->value) != NULL) return 1;
+                } else
+                {
+                    if(strstr(value,l->value) != NULL) return 1;
+                }
+                break;
+#ifdef HAVE_REGEX
+            case OP_REQEXP:
+                if(regexec(&l->reg,value,(size_t) 0, NULL, 0) == 0) return 1;
+                break;
+#endif
+        }
+        l = l->next;
+    }
+    return 0;
+}
+
+static int
+full_scan_expression(struct expression *e,char *value,int casecmp)
+{
+    register int i = 0;
+    struct expr_list *l;
+
+
+    if(e->fast_entries)
+    {
+        while(i < e->fast_entries)
+        {
+            if(scan_expression_list(e->expr_hash[e->fast_expr_hash[i]],value,e->op,casecmp)) return 1;
+            i++;
+        }
+    } else
+    {
+        while(i < MAX_EXPR_HASH)
+        {
+            if(scan_expression_list(e->expr_hash[i],value,e->op,casecmp)) return 1;
+            i++;
+        }
+    }
+    return 0;
+} 
+
 /* returns true if and = 0 and attleast one expression is true 
  * or and = 1 and all expressions are true
  */
 int
-eval_expression(struct structure *s, int and,int invert, uint8_t *buffer)
+eval_expression(struct structure *s,struct record *r, int and,int invert, int casecmp, uint8_t *buffer)
 {
     struct expression *e = expression;
     struct expr_list *el;
@@ -1886,15 +2169,28 @@ eval_expression(struct structure *s, int and,int invert, uint8_t *buffer)
     int prev_retval;
     int loop_break = 0;
     int expression_count = 0;
+    struct output *o;
+    size_t len;
 
     if(e == NULL) return 0;
 
-    e = expression;
-
     while(e != NULL && !loop_break)
     {
         if(e->f != NULL)
         {
+
+            o = e->f->o ? e->f->o : r->o;
+            if(o != no_output && o->hex_cap)
+            {
+                bcd_to_ascii = bcd_to_ascii_cap;
+                hex_to_ascii = hex_to_ascii_cap;
+            } else
+            {
+                bcd_to_ascii = bcd_to_ascii_low;
+                hex_to_ascii = hex_to_ascii_low;
+            }
+
+
             start_write();
             switch(s->type[0])
             {
@@ -1909,60 +2205,31 @@ eval_expression(struct structure *s, int and,int invert, uint8_t *buffer)
                     break;
             }
             writec(0);  // end of string
-            if(e->el != NULL)
-            {
-                el = e->el;
-                prev_retval = retval;
-                eq_found = 0;
+            prev_retval = retval;
+            eq_found = 0;
 
-                do
-                {
-                    switch(e->op)
-                    {
-                        case OP_START:
-                            if(strncmp(el->value,write_buffer,el->value_len) == 0) retval++;
-                            break;
-                        case OP_EQUAL:
-                            if(strcmp(el->value,write_buffer) == 0) retval++;
-                            break;
-                        case OP_CONTAINS:
-                            if(strstr(write_buffer,el->value) != NULL) retval++;
-                            break;
-                        case OP_NOT_EQUAL:
-                            if(strcmp(write_buffer,el->value) == 0) eq_found++;
-                            break;
-#ifdef HAVE_REGEX
-                        case OP_REQEXP:
-                            if(regexec(&el->reg,write_buffer,(size_t) 0, NULL, 0) == 0) retval++;
-                            break;
-#endif
-                    }
-                    el = el->next;
-                } while(el != NULL && prev_retval == retval && !eq_found);
-                if(e->op == OP_NOT_EQUAL && !eq_found) retval++;
-            }  else
+            switch(e->op)
             {
-                switch(e->op)
-                {
-                    case OP_START:
-                        if(strncmp(e->value,write_buffer,e->value_len) == 0) retval++;
-                        break;
-                    case OP_EQUAL:
-                        if(strcmp(e->value,write_buffer) == 0) retval++;
-                        break;
-                    case OP_CONTAINS:
-                        if(strstr(write_buffer,e->value) != NULL) retval++;
-                        break;
-                    case OP_NOT_EQUAL:
-                        if(strcmp(write_buffer,e->value) != 0) retval++;
-                        break;
-#ifdef HAVE_REGEX
-                    case OP_REQEXP:
-                        if(regexec(&e->reg,write_buffer,(size_t) 0, NULL, 0) == 0) retval++;
-                        break;
-#endif
-                }
-            }
+                case OP_START:
+                    len = e->exp_max_len;
+                    prev_retval = retval;
+                    while(len >= e->exp_min_len && prev_retval == retval)
+                    {
+                        retval += hash_scan_expression_len(e->expr_hash,write_buffer,casecmp,len);
+                        len--;
+                    }
+                    break;
+                case OP_CONTAINS:
+                case OP_REQEXP:
+                    retval += full_scan_expression(e,write_buffer,casecmp);
+                    break;
+                case OP_EQUAL:
+                    retval += hash_scan_expression(e->expr_hash,write_buffer,casecmp);
+                    break;
+                case OP_NOT_EQUAL:
+                    if(hash_scan_expression(e->expr_hash,write_buffer,casecmp) == 0) retval++;
+                    break;
+            } 
         }
         if(!and && retval)
         {
@@ -1988,14 +2255,14 @@ eval_expression(struct structure *s, int and,int invert, uint8_t *buffer)
     if(invert) retval = !retval;
     return retval;
 }
-
 /* initialize expressions field pointers */
 /* pointers are preinitialized for faster execution */
-void
+    void
 init_expression_list(struct record *r)
 {
     struct expression *e;
     struct field *f = r->f;
+    int exp_count = 0;
 
     e = expression;
 
@@ -2003,18 +2270,20 @@ init_expression_list(struct record *r)
     {
         e->f = NULL;
         e = e->next;
+        exp_count++;
     }
 
     e = expression;
 
-    while(f != NULL)
+    while(f != NULL && exp_count)
     {
         e = expression;
-        while(e != NULL)
+        while(e != NULL && exp_count)
         {
-            if(f->exp == e && e->f == NULL)
+            if(e->f == NULL && strcasecmp(f->name,e->field) == 0)
             {
                 e->f = f;
+                exp_count--;
             }
             e = e->next;
         }
@@ -2022,7 +2291,7 @@ init_expression_list(struct record *r)
     }
 }
 
-void
+    void
 open_debug_file(int stype)
 {
     sprintf(debug_file,"ffe_error_%d.log",(int) getpid());
@@ -2034,7 +2303,7 @@ open_debug_file(int stype)
         debug_fp = xfopen(debug_file,"w");
     }
 }
-    
+
 
 void
 write_debug_file(uint8_t *line,int len,int stype)
@@ -2052,11 +2321,24 @@ write_debug_file(uint8_t *line,int len,int stype)
     debug_lineno++;
 }
 
+static void
+select_output(struct output *o)
+{
+    if(o->ofp != NULL)
+    {
+        output_fp = o->ofp;
+        output_file = o->output_file;
+    } else
+    {
+        output_fp = default_output_fp;
+        output_file = default_output_file;
+    }
+}
 
 
 /* main loop for execution */
 void 
-execute(struct structure *s,int strict, int expression_and,int expression_invert,int debug)
+execute(struct structure *s,int strict, int expression_and,int expression_invert,int expression_case, int debug)
 {
     uint8_t *input_line;
     struct record *r = NULL;
@@ -2075,6 +2357,8 @@ execute(struct structure *s,int strict, int expression_and,int expression_invert
     i = 0;
     while(i < JUSTIFY_STRING) justify_string[i++] = ' ';
 
+    reset_levels(0,MAXLEVEL);
+
     write_buffer = xmalloc(write_buffer_size);
     write_buffer_end = write_buffer + (write_buffer_size - 1);
 
@@ -2084,6 +2368,7 @@ execute(struct structure *s,int strict, int expression_and,int expression_invert
         read_buffer = xrealloc(read_buffer,read_buffer_size);
     }
 
+    select_output(s->o);
     print_text(s,NULL,s->o->file_header);
     while((input_line = get_input_line(&length,s->type[0])) != NULL)
     {
@@ -2116,24 +2401,27 @@ execute(struct structure *s,int strict, int expression_and,int expression_invert
                 if((r->pf == NULL && r->o->no_data == 1) || r->pf != NULL || r->o == raw)
                 {
                     update_field_positions(s->type,s->quote,r->arb_length,r->f,length,input_line);
-                    if(expression == NULL || (eval_expression(s,expression_and,expression_invert,input_line)))
+                    if(expression == NULL || (eval_expression(s,r,expression_and,expression_invert,expression_case,input_line)))
                     {
                         if(r->o == raw)
                         {
                             print_raw(s->type[0] == BINARY ? r->length : length,input_line,s->type[0]);
                         } else
                         {
-
+                            select_output(r->o);
+                            print_level_before(prev_record,r);
                             if(r->o->header != NULL && !header_printed) 
                             {
                                 header_printed = print_header(s,r);
                             }
-                            if(r->o->indent != NULL && r->o->record_header != NULL) print_text(s,r,r->o->indent);
+                            if(r->o->indent != NULL && r->o->record_header != NULL) 
+                                print_indent(r->o->indent,r->level != NULL ? r->level->level : 1);
                             print_text(s,r,r->o->record_header);
                             fields_printed = print_fields(s,r,input_line);
                             if(fields_printed || r->o->print_empty)
                             {
-                                if(r->o->indent != NULL && r->o->record_trailer != NULL) print_text(s,r,r->o->indent);
+                                if(r->o->indent != NULL && r->o->record_trailer != NULL) 
+                                    print_indent(r->o->indent,r->level != NULL ? r->level->level : 1);
                                 print_text(s,r,r->o->record_trailer);
                             }
                         }
@@ -2143,6 +2431,8 @@ execute(struct structure *s,int strict, int expression_and,int expression_invert
             if(first_line) first_line = 0;
         }
     }
+    print_level_end(r);
+    select_output(s->o);
     print_text(s,r,s->o->file_trailer);
     free(write_buffer);
     if(debug_fp != NULL) fclose(debug_fp);
