@@ -17,24 +17,29 @@
  *    You should have received a copy of the GNU General Public License
  *    along with ffe; if not, write to the Free Software
  *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ *    F607480034
+ *    HJ9004-2
  *
  */
 
-/* $Id: ffe.c,v 1.80 2008-07-11 14:13:02 timo Exp $ */
+/* $Id: ffe.c,v 1.88 2011-04-10 10:12:10 timo Exp $ */
 
 #include "ffe.h"
+#include <ctype.h>
+#include <stdlib.h>
 
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #endif
 
-#include <ctype.h>
-#include <stdlib.h>
-
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
 
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#endif
+
 #ifdef PACKAGE
 static char *program = PACKAGE;
 #else
@@ -53,13 +58,19 @@ static char *host = HOST;
 static char *host = "";
 #endif
 
+#ifdef BUILD_DATE
+static char *build_date = BUILD_DATE;
+#else
+static char *build_date = "";
+#endif
+
 #ifdef PACKAGE_BUGREPORT
 static char *email_address = PACKAGE_BUGREPORT;
 #else
 static char *email_address = "tjsa@iki.fi";
 #endif
 
-static char short_opts[] = "c:s:o:p:f:e:r:l?VavdI";
+static char short_opts[] = "c:s:o:p:f:e:r:l?VavdIX";
 
 #ifdef HAVE_GETOPT_LONG
 static struct option long_opts[] = {
@@ -77,6 +88,7 @@ static struct option long_opts[] = {
     {"replace",1,NULL,'r'},
     {"debug",0,NULL,'d'},
     {"info",0,NULL,'I'},
+    {"casecmp",0,NULL,'X'},
     {NULL,0,NULL,0}
 };
 #endif
@@ -87,6 +99,8 @@ extern struct input_file *files;
 
 int system_endianess = F_UNKNOWN_ENDIAN;
 int max_binary_record_length = 0;
+char *ffe_open = NULL;
+
 
 
 /* global rc-data */
@@ -180,10 +194,11 @@ get_default_rc_name()
 void
 help(FILE *stream)
 {
+    char *rc=get_default_rc_name();
     fprintf(stream,"Usage: %s [OPTION]...\n\n",program);
 #ifdef HAVE_GETOPT_LONG
     fprintf(stream,"-c, --configuration=FILE\n");
-    fprintf(stream,"\t\tRead configuration from FILE, default is \'%s\'.\n",get_default_rc_name());
+    fprintf(stream,"\t\tRead configuration from FILE, default is \'%s\'.\n",rc);
     fprintf(stream,"-s, --structure=STRUCTURE\n");
     fprintf(stream,"\t\tUse structure STRUCTURE for input file, suppresses guessing.\n");
     fprintf(stream,"-p, --print=FORMAT\n");
@@ -196,6 +211,8 @@ help(FILE *stream)
     fprintf(stream,"\t\tPrint only those records for which the EXPRESSION evaluates to true.\n");
     fprintf(stream,"-a, --and\n");
     fprintf(stream,"\t\tExpressions are combined with logical and, default is logical or.\n");
+    fprintf(stream,"-X, --casecmp\n");
+    fprintf(stream,"\t\tExpressions are evaluated case insensitive.\n");
     fprintf(stream,"-v, --invert-match\n");
     fprintf(stream,"\t\tPrint only those records which don't match the expression.\n");
     fprintf(stream,"-l, --loose\n");
@@ -211,7 +228,7 @@ help(FILE *stream)
     fprintf(stream,"-V, --version\n");
 #else
     fprintf(stream,"-c FILE\n");
-    fprintf(stream,"\t\tRead configuration from FILE, default is \'%s\'.\n",get_default_rc_name());
+    fprintf(stream,"\t\tRead configuration from FILE, default is \'%s\'.\n",rc);
     fprintf(stream,"-s STRUCTURE\n");
     fprintf(stream,"\t\tUse structure STRUCTURE for input file, suppresses guessing.\n");
     fprintf(stream,"-p FORMAT\n");
@@ -224,6 +241,8 @@ help(FILE *stream)
     fprintf(stream,"\t\tPrint only those records for which the EXPRESSION evaluates to true.\n");
     fprintf(stream,"-a\n");
     fprintf(stream,"\t\tExpressions are combined with logical and, default is logical or.\n");
+    fprintf(stream,"-X\n");
+    fprintf(stream,"\t\tExpressions are evaluated case insensitive.\n");
     fprintf(stream,"-v\n");
     fprintf(stream,"\t\tPrint only those records which don't match the expression.\n");
     fprintf(stream,"-l\n");
@@ -242,6 +261,7 @@ help(FILE *stream)
     fprintf(stream,"\nAll remaining arguments are names of input files;\n");
     fprintf(stream,"if no input files are specified, then the standard input is read.\n");
     fprintf(stream,"\nSend bug reports to %s.\n",email_address);
+    free(rc);
 }
 
 void
@@ -254,12 +274,36 @@ usage(int opt)
 void
 print_version()
 {
-    printf("%s version %s (%s)\n",program,version,host);
+    printf("%s version %s\n%s %s\n",program,version,build_date,host);
     printf("Copyright (c) 2007 Timo Savinen\n\n");
     printf("This is free software; see the source for copying conditions.\n");
     printf("There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
 }
 
+size_t
+hash(char *str,size_t len)
+{
+    register unsigned long h = 5381;
+    int c;
+
+    if(len > 0)
+    {
+        while ((c = *str++) != 0 && len > 0)
+        {
+            h = ((h << 5) + h) + c;
+            len--;
+        }
+    }else
+    {
+        while ((c = *str++) != 0)
+        {
+            h = ((h << 5) + h) + c;
+        }
+    }
+    return (size_t) (h % MAX_EXPR_HASH);
+}
+
+
 
 struct output *
 search_output(char *name)
@@ -376,6 +420,7 @@ check_rc(char *use_output)
         {
             if(r->output_name == NULL || use_output != NULL) 
             {
+                if(r->output_name != NULL) free(r->output_name);
                 r->output_name = s->output_name;
                 r->o = s->o;
             } else
@@ -516,6 +561,18 @@ check_rc(char *use_output)
     while(o != NULL)
     {
         if(o->lookup == NULL) o->lookup = o->data;
+        if(o->output_file != NULL)
+        {
+            if(o->output_file[0] == '-' && o->output_file[1] == 0)
+            {
+                o->ofp = stdout;
+                free(o->output_file);
+                o->output_file = "(stdout)";
+            } else
+            {
+                o->ofp = xfopen(o->output_file,"w");
+            }
+        }
         o = o->next;
     }
 
@@ -559,18 +616,39 @@ add_replace(char *optarg)
     r->found = 0;
 }
 
-struct expr_list *
-read_expression_file(char *file)
+
+/* adds an item to expression hash list 
+    */
+static void
+add_expression_to_list(struct expr_list **list, char *value)
+{
+    size_t h = hash(value,0);
+    register struct expr_list *e = list[h];
+
+    if(e == NULL)
+    {
+        e = xmalloc(sizeof(struct expr_list));
+        list[h] = e;
+    } else
+    {
+        while(e->next != NULL) e = e->next;
+        e->next = xmalloc(sizeof(struct expr_list));
+        e = e->next;
+    }
+    e->value = xstrdup(value);
+    e->value_len = strlen(e->value);
+    e->next = NULL;
+}
+
+
+static void
+read_expression_file(struct expr_list **list, char *file)
 {
-    struct expr_list *retval,*e;
     FILE *fp;
-    int ccount;
+    register int ccount;
     size_t line_len = 1024;
     char *line = xmalloc(line_len);
 
-    e = NULL;
-    retval = NULL;
-
     fp = xfopen(file,"r");
 
     do
@@ -586,29 +664,23 @@ read_expression_file(char *file)
             ccount = -1;
         }
 #endif
-        if (ccount != -1)
+        if (ccount > 1)
         {
-            if(ccount) ccount--;
-            line[ccount] = 0;
-            if(e == NULL)
-            {
-                e = xmalloc(sizeof(struct expr_list));
-                e->value = xstrdup(line);
-                retval = e;
-            } else
-            {
-                e->next = xmalloc(sizeof(struct expr_list));;
-                e = e->next;
-                e->value = xstrdup(line);
-            }
-            e->value_len = ccount;
-            e->next = NULL;
+            line[ccount - 1] = 0;
+            add_expression_to_list(list,line);
         }
     }
     while(ccount != -1);
     fclose(fp);
     free(line);
-    return retval;
+}
+
+static void
+init_expr_hash(struct expr_list **list)
+{
+    register int i = 0;
+
+    while(i < MAX_EXPR_HASH) list[i++] = NULL;
 }
 
 
@@ -617,11 +689,10 @@ add_expression(char *optarg)
 {
     char *op_pos;
     char op = 0;
-    struct expression *e;
+    struct expression *e,*last;
     struct expr_list *el;
-    int rc;
+    int found = 0;
     size_t buflen;
-    char *errbuf;
     char *value_file;
 
     if((op_pos = strchr(optarg,OP_REQEXP)) != NULL)
@@ -657,61 +728,99 @@ add_expression(char *optarg)
         expression = xmalloc(sizeof(struct expression));
         expression->next =  NULL;
         e = expression;
+        e->field = xstrdup(optarg);
+        e->exp_min_len = 0;
+        e->exp_max_len = 0;
+        e->fast_entries = 0;
+        init_expr_hash(e->expr_hash); 
     } else
     {
-        while(e->next != NULL) e = e->next;
-        e->next = xmalloc(sizeof(struct expression));
-        e = e->next;
-        e->next = NULL;
+        do
+        {
+            if(strcasecmp(optarg,e->field) == 0 && e->op == op) 
+            {
+                found = 1;
+            } else
+            {
+                last = e;
+                e = e->next;
+            }
+        }
+        while(e != NULL && !found);
+
+        if(!found)
+        {
+            last->next = xmalloc(sizeof(struct expression));
+            e = last->next;
+            e->next = NULL;
+            e->field = xstrdup(optarg);
+            e->exp_min_len = 0;
+            e->exp_max_len = 0;
+            e->fast_entries = 0;
+            init_expr_hash(e->expr_hash);
+        }
     }
 
-    e->field = xstrdup(optarg);
     op_pos++;
     if(strstr(op_pos,"file:") == op_pos)
     {
         value_file = expand_home(&op_pos[5]);
-        e->el = read_expression_file(value_file);
-        if(e->el == NULL) panic("No values in file",value_file,NULL);
+        read_expression_file(e->expr_hash,value_file);
         free(value_file);
     } else
     {
-        e->value = xstrdup(op_pos);
-        e->value_len = strlen(e->value);
-        e->el = NULL;
+        add_expression_to_list(e->expr_hash,op_pos);
     }
     e->found = 0;
     e->op = op;
-#ifdef HAVE_REGEX
-    if(e->op == OP_REQEXP)
+}
+
+static void
+init_expression()
+{
+    struct expression *e = expression;
+    register struct expr_list *l;
+    register int i;
+    int rc,min_init;
+    size_t buflen;
+    char *errbuf;
+
+    while(e != NULL)
     {
-        if(e->el != NULL)
+        i = 0;
+        min_init = 0;
+        while(i < MAX_EXPR_HASH)
         {
-            el = e->el;
-            while(el != NULL)
+            l = e->expr_hash[i];
+
+	        if (l != NULL && e->fast_entries <= MAX_EXPR_FAST_LIST) e->fast_expr_hash[e->fast_entries++] = i; 
+
+	        while(l != NULL)
             {
-                rc = regcomp(&el->reg,el->value,REG_EXTENDED | REG_NOSUB);
-                if(rc)
+                if(!min_init) {min_init = 1;e->exp_min_len = l->value_len;}
+                if(l->value_len < e->exp_min_len) e->exp_min_len = l->value_len;
+                if(l->value_len > e->exp_max_len) e->exp_max_len = l->value_len;
+
+#ifdef HAVE_REGEX
+		        if(e->op == OP_REQEXP)
                 {
-                    buflen = regerror(rc,&e->reg,NULL,0);
-                    errbuf = xmalloc(buflen + 1);
-                    regerror(rc,&e->reg,errbuf,buflen);
-                    panic("Error in regular expression",el->value,errbuf);
+                    rc = regcomp(&l->reg,l->value,REG_EXTENDED | REG_NOSUB);
+                    if(rc)
+                    {
+                        buflen = regerror(rc,&l->reg,NULL,0);
+                        errbuf = xmalloc(buflen + 1);
+                        regerror(rc,&l->reg,errbuf,buflen);
+                        panic("Error in regular expression",l->value,errbuf);
+                    }
                 }
-                el = el->next;
-            }
-        } else
-        {
-            rc = regcomp(&e->reg,e->value,REG_EXTENDED | REG_NOSUB);
-            if(rc)
-            {
-                buflen = regerror(rc,&e->reg,NULL,0);
-                errbuf = xmalloc(buflen + 1);
-                regerror(rc,&e->reg,errbuf,buflen);
-                panic("Error in regular expression",e->value,errbuf);
-            }
-        }
-    }
 #endif
+                l = l->next;
+            }
+            i++;
+        }
+        if(e->fast_entries > MAX_EXPR_FAST_LIST) e->fast_entries = 0; 
+        e = e->next;
+    }
 }
 
 /* if variable is set it will not be overwritten */
@@ -769,7 +878,7 @@ environment(char *structure, char *format,char *ofile)
     free(file_names);
 }
 
-    
+
 
 int
 main(int argc, char **argv)
@@ -780,12 +889,22 @@ main(int argc, char **argv)
     int info = 0;
     int expression_and = 0;
     int expression_invert = 0;
+    int expression_casecmp = 0;
     struct structure *s = NULL;
     char *structure_to_use = NULL;
     char *output_to_use = NULL;
     char *config_to_use = NULL;
     char *ofile_to_use = NULL;
     char *field_list = NULL;
+
+#ifdef HAVE_SIGACTION
+    struct sigaction act;
+    sigemptyset(&act.sa_mask);
+    act.sa_handler = SIG_IGN;
+    act.sa_flags = SA_NOCLDWAIT;
+    sigaction (SIGCHLD, &act, NULL);
+#endif
+
 #ifdef HAVE_GETOPT_LONG
     while ((opt = getopt_long(argc,argv,short_opts,long_opts,NULL)) != -1)
 #else
@@ -848,6 +967,9 @@ main(int argc, char **argv)
                 case 'a':
                     expression_and = 1;
                     break;
+                case 'X':
+                    expression_casecmp = 1;
+                    break;
                 case 'd':
                     debug = 1;
                     break;
@@ -893,12 +1015,16 @@ main(int argc, char **argv)
     parserc(config_to_use,field_list);
 
     check_rc(output_to_use);
+     
+    init_expression();
 
     if(info)
     {
         print_info();
         exit(EXIT_SUCCESS);
     }
+    
+    ffe_open = getenv("FFEOPEN");
 
     if(structure_to_use == NULL)
     {
@@ -922,9 +1048,11 @@ main(int argc, char **argv)
         }
     }
 
+    free(config_to_use); /* to avoid strange valgrind memory lost */
+
     set_output_file(ofile_to_use);
 
-    execute(s,strict,expression_and,expression_invert,debug);
+    execute(s,strict,expression_and,expression_invert,expression_casecmp,debug);
 
     close_output_file();
 
