@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: parserc.c,v 1.76 2008-09-11 06:44:21 timo Exp $ */
+/* $Id: parserc.c,v 1.80 2010-09-15 10:49:54 timo Exp $ */
 /* parsing the rc-file */
 
 #include "ffe.h"
@@ -103,6 +103,10 @@ char *values[100];            /* poister to option name and parameters */
 #define N_SEPARATOR         "separator"
 #define N_RECORD_HEADER     "record-header"
 #define N_RECORD_TRAILER    "record-trailer"
+#define N_GROUP_HEADER      "group-header"
+#define N_ELEMENT_HEADER    "element-header"
+#define N_GROUP_TRAILER     "group-trailer"
+#define N_ELEMENT_TRAILER   "element-trailer"
 #define N_JUSTIFY           "justify"
 #define N_INDENT            "indent"
 #define N_FIELDLIST         "field-list"
@@ -116,6 +120,10 @@ char *values[100];            /* poister to option name and parameters */
 #define N_SEARCH            "search"
 #define N_CONST             "const"
 #define N_FIELD_COUNT       "field-count"
+#define N_OFILE             "output-file"
+#define N_LEVEL             "level"
+#define N_RECORD_LENGTH     "record-length"
+#define N_HEX_CAP           "hex-caps"
 
 
 
@@ -137,6 +145,10 @@ static struct rc_option rc_opts[] = {
     {N_SEPARATOR,"S"},
     {N_RECORD_HEADER,"S"},
     {N_RECORD_TRAILER,"S"},
+    {N_GROUP_HEADER,"S"},
+    {N_GROUP_TRAILER,"S"},
+    {N_ELEMENT_HEADER,"S"},
+    {N_ELEMENT_TRAILER,"S"},
     {N_JUSTIFY,"S"},
     {N_INDENT,"S"},
     {N_FIELDLIST,"S"},
@@ -150,6 +162,10 @@ static struct rc_option rc_opts[] = {
     {N_SEARCH,"S"},
     {N_CONST,"SS"},
     {N_FIELD_COUNT,"N"},
+    {N_OFILE,"S"},
+    {N_LEVEL,"Nss"},
+    {N_RECORD_LENGTH,"S"},
+    {N_HEX_CAP,"S"},
     {NULL,NULL}
 };
 
@@ -279,7 +295,7 @@ execute_child(char *command)
     {
         close(fds[0]);
         if(dup2(fds[1],STDOUT_FILENO) == -1) panic("dup2 error",strerror(errno),NULL);
-        if(system(command) == -1) panic("Command failed",command,strerror(errno));
+        if(execl(SHELL_CMD, "sh", "-c", command, NULL) == -1) panic("Starting a shell with execl failed",command,strerror(errno));
         close(fds[1]);
         _exit(EXIT_SUCCESS);
     } else if(pid > (pid_t) 0)
@@ -1217,11 +1233,18 @@ parserc(char *rcfile,char *include_field_list)
                             c_output->separator = NULL;
                             c_output->record_header = NULL;
                             c_output->record_trailer = "\n";
+                            c_output->group_header = NULL;
+                            c_output->group_trailer = NULL;
+                            c_output->element_header = NULL;
+                            c_output->element_trailer = NULL;
                             c_output->justify = LEFT_JUSTIFY;
                             c_output->indent = NULL;
                             c_output->no_data = 1;
+                            c_output->hex_cap = 0;
                             c_output->empty_chars = " \f\n\r\t\v";
                             c_output->print_empty = 1;
+                            c_output->output_file = NULL;
+                            c_output->ofp = NULL;
                             if(fl != NULL)
                             {
                                 c_output->fl = fl;
@@ -1359,7 +1382,8 @@ parserc(char *rcfile,char *include_field_list)
                             c_record->o = NULL;
                             c_record->output_name = NULL;
                             c_record->vote = 0;
-                            c_record->arb_length = 0;
+                            c_record->arb_length = RL_STRICT;
+                            c_record->level = NULL;
                             status = PS_W_RECORD;
                         } else if(strcmp(values[0],N_QUOTE) == 0)
                         {
@@ -1439,7 +1463,6 @@ parserc(char *rcfile,char *include_field_list)
                             c_field->type = F_ASC;
                             c_field->next = NULL;
                             c_field->const_data = NULL;
-                            c_field->exp = NULL;
                             c_field->length = 0;
                             c_field->output_name = NULL;
                             c_field->o = NULL;
@@ -1500,12 +1523,46 @@ parserc(char *rcfile,char *include_field_list)
                                 c_field->rep = NULL;
                                 c_field->next = NULL;
                                 c_field->const_data = NULL;
-                                c_field->exp = NULL;
                                 c_field->length = 0;
                                 c_field->name = NULL;
                                 c_field->output_name = NULL;
                                 c_field->o = NULL;
                             }
+                        } else if(strcmp(values[0],N_LEVEL) == 0)
+                        {
+                            c_record->level = xmalloc(sizeof(struct level));
+                            c_record->level->element_name = NULL;
+                            c_record->level->group_name = NULL;
+                            c_record->level->level = atoi(values[1]);
+                            c_record->level->indent_count = 0;
+                            if (c_record->level->level < 1 || c_record->level->level > MAXLEVEL)
+                            {
+                                error_in_line();
+                                panic("Invalid level value",NULL,NULL);
+                            }
+                            if(opt_count > 1 && strcmp(values[2],"*") != 0)
+                            {
+                                c_record->level->element_name = xstrdup(values[2]);
+                                c_record->level->indent_count++;
+                            }
+                            if(opt_count > 2)
+                            {
+                                c_record->level->group_name = xstrdup(values[3]);
+                                c_record->level->indent_count++;
+                            }
+                        } else if(strcmp(values[0],N_RECORD_LENGTH) == 0)
+                        {
+                            if(strcmp(values[1],"strict") == 0)
+                            {
+                                c_record->arb_length = RL_STRICT;
+                            } else if(strcmp(values[1],"minimum") == 0)
+                            {
+                                c_record->arb_length = RL_MIN;
+                            } else
+                            {
+                                error_in_line();
+                                panic("Unknown option value in record",NULL,NULL);
+                            }
                         } else
                         {
                             error_in_line();
@@ -1536,7 +1593,25 @@ parserc(char *rcfile,char *include_field_list)
                             c_output->record_header = xstrdup(values[1]);
                         } else if(strcmp(values[0],N_RECORD_TRAILER) == 0)
                         {
-                            c_output->record_trailer = xstrdup(values[1]);
+                            if (values[1][0])
+                            {
+                                c_output->record_trailer = xstrdup(values[1]);
+                            } else
+                            {
+                                c_output->record_trailer = NULL;
+                            }
+                        } else if(strcmp(values[0],N_GROUP_HEADER) == 0)
+                        {
+                            c_output->group_header = xstrdup(values[1]);
+                        } else if(strcmp(values[0],N_GROUP_TRAILER) == 0)
+                        {
+                            c_output->group_trailer = xstrdup(values[1]);
+                        } else if(strcmp(values[0],N_ELEMENT_HEADER) == 0)
+                        {
+                            c_output->element_header = xstrdup(values[1]);
+                        } else if(strcmp(values[0],N_ELEMENT_TRAILER) == 0)
+                        {
+                            c_output->element_trailer = xstrdup(values[1]);
                         } else if(strcmp(values[0],N_JUSTIFY) == 0)
                         {
                             if(!values[1][1])
@@ -1588,7 +1663,23 @@ parserc(char *rcfile,char *include_field_list)
                         } else if(strcmp(values[0],N_EMPTY_CHARS) == 0)
                         {
                             c_output->empty_chars = xstrdup(values[1]);
-                        } else
+                        } else if(strcmp(values[0],N_OFILE) == 0)
+                        {
+                            c_output->output_file = xstrdup(values[1]);
+                        } else if(strcmp(values[0],N_HEX_CAP) == 0)
+                        {
+                            if(strcmp(values[1],"yes") == 0)
+                            {
+                                c_output->hex_cap = 1;
+                            } else if(strcmp(values[1],"no") == 0)
+                            {
+                                c_output->hex_cap = 0;
+                            } else
+                            {
+                                error_in_line();
+                                panic("Unknown values in hex_cap",NULL,NULL);
+                            }
+                        } else 
                         {
                             error_in_line();
                             panic("Unknown option in output definition",NULL,NULL);
